
/* wind volume reader test. 
	Counts appear on gpio22 
	When one is received, this is 2/3 m
	to be added to a register.

	Somewhat like the accumulator 
*/


/* GPIO22 pin P1-15 2/3 m  pulses input, react to the
	falling edge.
*/

/* Signal SIG_CLWCV to be sent if so enabled upon each CL 
	pulse received: */

#define SIG_CLWCV SIGRTMIN+3


#include <linux/init.h>
#include <linux/miscdevice.h>

#include <linux/module.h>
#include <linux/kernel.h> /* printk() */

#include <linux/interrupt.h>
#include <linux/gpio.h>

#include <linux/signal.h>  /* Signal numbers etc. */
#include <linux/sched.h>  /* process-ident things. */
#include <linux/sched/signal.h>  /* send_sig_info . */

#include <linux/slab.h>   /* kmalloc() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/proc_fs.h>
#include <linux/fcntl.h>  /* O_ACCMODE */

#include <linux/uaccess.h>  /* copy_from/to_user */

#include "windvol_u.h"  /* Signal numbers. */


MODULE_LICENSE("Dual BSD/GPL");

#define FCMBUFSIZE 100

#define CL_LABEL "wind volume"

#define GPIO_PIN_DEV_STR "WINDVOL"
#define THISFILE "windvol.c"

/* Declaration of windvol.c functions */

static int f_open(struct inode *inode, struct file *filp);
static int f_release(struct inode *inode, struct file *filp);
static ssize_t f_read(struct file *filp, char *buf, size_t count, loff_t *f_pos);
static ssize_t f_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos);
static void f_exit(void);
static int f_init(void);

/* Structure that declares the access functions. */

struct file_operations windvol_fops = 
{
  read: f_read,
  write: f_write,
  open: f_open,
  release: f_release
};

/* Misc device structure. */
static struct miscdevice windvol_misc_device 
	= {
	.minor = MISC_DYNAMIC_MINOR,
	.name = "windvol",
	.fops = &windvol_fops 
  };

/* Declaration of the init and exit functions */
module_init(f_init);
module_exit(f_exit);

/* enforce opening only once. */
static int f_open_count = 0;

/* Which signal to send and where to send it? */
pid_t send_to_pid = 0; /* Set in write or better, ioctl. */

short int cl_signal = SIG_CLWCV;
module_param(cl_signal, short, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP ); 
MODULE_PARM_DESC(cl_signal, "Signal to send on each 2/3 m"); 


/* Global variables of the driver. */

static short int windvol_cl_irq  = 0;
static short int windvol_irq_enabled;

static unsigned int windvol_total;
static int windvol_countdown;


static short gpio_cl = 22;


/* Send the indicated signal if the send_to_pid is nonzero and oke */ 
int sendasignal(int id, short sigtosend)
{
	int ssig;
	struct task_struct *task_from_pid;

	printk(KERN_INFO "%s(%d): %c: About to send signal %d to %d\n", THISFILE, __LINE__, id, sigtosend, send_to_pid); 

	/* The following sends the signal to the 
		listening process, if there is one. */ 
	if(send_to_pid)
	{
		task_from_pid = pid_task(find_vpid(send_to_pid), PIDTYPE_PID); 

		if(task_from_pid == NULL) 
		{
			// no-one to send a signal to
			return(0);
		}

		ssig = send_sig_info(sigtosend, SEND_SIG_FORCED, task_from_pid);
		if(ssig < 0)
		{
			// signal wasn't sent
			return(0);
		}

		// signal has been sent
		return(ssig);
	}
	// signal not sent as there were noone to send it to 
	return(0);
}


/* IRQ handler */

/* CL pulse is detected.
	Increment the totalizer
	Decrement the countdown, if 0 turn the relay off.
	If there is a process, send the SIG_CLWCV signal to a user process.	
*/

static irqreturn_t f_irq_cl_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned long flags;
	

	local_irq_save(flags); 

	if(!windvol_irq_enabled)
	{
		// printk(KERN_INFO "%s(%d): Irq not active\n", THISFILE, __LINE__); 
		goto theirqisdone;
	}

	windvol_total++;
	windvol_countdown--;
	// if(windvol_countdown == 0)
	// {
		// gpio_set_value(gpio_relay, FCRELAYOFF);
	// }

	/* Never go below 0. */ 
	if(windvol_countdown < 0) windvol_countdown = 0; 

	sendasignal('c', cl_signal); 

theirqisdone:

	local_irq_restore(flags); 
	return IRQ_HANDLED;
}


/* use:
	gpio_direction_input(gpio_pb);
	value = gpio_get_value(gpio_pb);

	or: 
	gpio_direction_output(gpio_pb, value);
	gpio_set_value(gpio_pb, value);
*/ 


/* Init: enable interrupt, create the device entry. */
static int f_init(void) 
{
	int rx;

	/* Request the input gpio line and enable the interrupts on them. */
	rx = gpio_request(gpio_cl, CL_LABEL);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d request failure %d\n", 
			THISFILE, __LINE__, gpio_cl, rx); 
		goto noclgpio;
	}

	rx = gpio_direction_input(gpio_cl);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d input direction request failure %d\n", 
			THISFILE, __LINE__, gpio_cl, rx); 
		goto noclinput;
	}

#if(0)
	rx = gpio_request( gpio_relay, RELAY_LABEL);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d request failure %d\n", 
			THISFILE, __LINE__, gpio_relay, rx); 
		goto norelaygpio;
	}

	rx = gpio_direction_output(gpio_relay, FCRELAYOFF);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d output direction request failure %d\n", 
			THISFILE, __LINE__, gpio_relay, rx); 
		goto norelayoutput;
	}

	gpio_set_value(gpio_relay, FCRELAYOFF);
#endif


	/* Get the IRQ for the CL line */ 

	/* Set the irqs */
	windvol_cl_irq = gpio_to_irq(gpio_cl);
	if(windvol_cl_irq < 0)
	{
		printk(KERN_INFO "%s(%d): GPIO %d to IRQ mapping failure %d\n", 
			THISFILE, __LINE__, gpio_cl, windvol_cl_irq); 
		rx = windvol_cl_irq;
		goto noclirq;
	}

	printk(KERN_INFO "%s(%d): Mapped int %d from gpio-line %d\n", 
			THISFILE, __LINE__, windvol_cl_irq, gpio_cl); 

	rx = request_irq(windvol_cl_irq, 
			(irq_handler_t) f_irq_cl_handler, 
			IRQF_TRIGGER_FALLING, /* IRQF_TRIGGER_RISING is an alternative */
			CL_LABEL,
			GPIO_PIN_DEV_STR); 

	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): Irq %d trigger-request on gpio %d failure %d\n", THISFILE, __LINE__, windvol_cl_irq, gpio_cl, rx); 
		goto noclirq;
	}

	printk(KERN_INFO "%s(%d): %s: gpio-line %d\n", THISFILE, __LINE__, CL_LABEL, gpio_cl); 
	// printk(KERN_INFO "%s(%d): %s: gpio-line %d\n", THISFILE, __LINE__, RELAY_LABEL, gpio_relay); 


	/* Registering the device itself. */
	rx = misc_register(&windvol_misc_device);
	if(rx < 0)
	{
		printk(KERN_INFO "%s(%d): Register misc device failure %d\n", THISFILE, __LINE__, rx); 
		goto nodevice;
	}

	printk(KERN_INFO "%s(%d): Info: misc device minor number = %d\n", 
		THISFILE, __LINE__, windvol_misc_device.minor);

	return 0;


/* Error exits */
nodevice:

	free_irq(windvol_cl_irq, GPIO_PIN_DEV_STR); 
noclirq:

// norelayoutput:
	// gpio_free(gpio_relay); 
// norelaygpio:
noclinput:
	gpio_free(gpio_cl); 
noclgpio:
	return(rx);
}


static void f_exit(void) 
{
	free_irq(windvol_cl_irq, GPIO_PIN_DEV_STR); 
	// gpio_free(gpio_relay); 
	gpio_free(gpio_cl); 

	misc_deregister(&windvol_misc_device);
	printk(KERN_INFO "%s(%d): Removing windvol\n", THISFILE, __LINE__);
}


static int f_open(struct inode *inode, struct file *filp) 
{
	if(f_open_count)
		return -EBUSY;

	f_open_count ++;
	try_module_get(THIS_MODULE); 

	return 0;
}

static int f_release(struct inode *inode, struct file *filp) 
{
	f_open_count --;
	module_put(THIS_MODULE);
	return 0;
}


static ssize_t f_read(struct file *filp, char *buff_out, size_t count, loff_t *f_pos) 
{ 
	size_t wcount;
	char wbuff[FCMBUFSIZE];

	// int relayvalue;
	int cuu;

	/* Read: return values of the various statuses:
		Total
		Countdown
		Relay state
		intr setting
	*/
	// relayvalue = gpio_get_value(gpio_relay);

	snprintf(wbuff, sizeof(wbuff), 
		"T:%u C:%d I:%d P:%d\n", 
		windvol_total, windvol_countdown, 
		windvol_irq_enabled, send_to_pid); 

	wcount = strlen(wbuff);

	/* Transfering data to user space */ 
	cuu = copy_to_user(buff_out, wbuff, wcount);

	/* Changing reading position as best suits */ 
	if(*f_pos == 0) 
	{ 
		*f_pos += wcount; 
		return wcount; 
	} 
	else 
	{ 
		return 0; 
	}
}


/* Write -- we have to come in with some action from the outside,

	commands:
		E - enable the irq
		D - disable the irq
		Pnnnn - indicate process where to send signals, nnnn is a process id, pid 
		T - reset total
		Cnnn - set value nnn to the countdown
		N - Relay on
		F - Relay off
*/

static ssize_t f_write(struct file *filp, const char *buff_in, size_t count, loff_t *f_pos) 
{
	char rdbuff[FCMBUFSIZE];
	int skf;
	char *cp;
	int cuu;

	if(count > FCMBUFSIZE)
	{
		printk(KERN_INFO "%s(%d): Read error buffer overflow %zu > %u\n", THISFILE, __LINE__, count, FCMBUFSIZE); 
		return(-EINVAL);
	}
	
	cuu = copy_from_user(rdbuff, buff_in, count);
	if(cuu < 0)
	{	
		printk(KERN_INFO "%s(%d): Read error buffer underflow %d < %zu\n", THISFILE, __LINE__, cuu, count); 
		return(-EINVAL);
	}

	rdbuff[count] = 0;
	if(rdbuff[count-1] == '\n') rdbuff[count-1] = 0; 

	rcu_read_unlock(); /* Must use or code will loop ? */ 

	for(cp = rdbuff; *cp; cp++)
	{
		if(*cp == 'E') /* E - enable the irq */ 
		{
			windvol_irq_enabled = 1; 
		}
		else
		if(*cp == 'D') /* D - disable the irq */ 
		{
			windvol_irq_enabled = 0; 
		}
		else
		if(*cp == 'T') /* T - reset total */ 
		{
			windvol_total = 0;
		}
		else
		if(*cp == 'N') /* N - Relay on */ 
		{
			// gpio_set_value(gpio_relay, FCRELAYON);
		}
		else
		if(*cp == 'F') /* F - Relay off */ 
		{
			// gpio_set_value(gpio_relay, FCRELAYOFF);
		}
		else
		if(*cp == 'P') /* Pnnnn - indicate process where to send signals, nnnn is a process id, pid  */ 
		{
			skf = sscanf(cp+1, "%d", &send_to_pid);
			if(skf == 1)
			{
				windvol_irq_enabled = 1; 
			}
			else /* sscanf failed or did something else unwanted */
			{
				send_to_pid = 0; 
				windvol_irq_enabled = 0; 
			}
		}
		else
		if(*cp == 'C') /* Cnnn - set value nnn to the countdown */ 
		{
			skf = sscanf(cp+1, "%d", &windvol_countdown);
		}
	}

	return(count); 
}



