
/* wreader.c -- user-space part, 
	receive the signal for wind pulses, and react to them. */ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


#include "windvol_u.h"

/* Read some file in /sys or /dev */ 
int WVOL_ReadFile(char *filename, char *txt, size_t txtmax)
{
	FILE *fp;
	char *fgp;

	memset(txt, 0, txtmax); 

	fp = fopen(filename, "rt");
	if(fp == NULL)
	{
		perror(filename);
		return(-1);
	}

	fgp = fgets(txt, txtmax, fp);
	if(fgp == NULL)
	{
		memset(txt, 0, txtmax); 
	}
	fclose(fp);
	return(0);
}

/* Read the /sys/module/windvol/parameters/cl_signal pseudo-file
	to determine which signal we are using. */ 

int WVOL_FindSignal()
{
	int sa, fs;
	char txt[30];

	fs = WVOL_ReadFile(SIGINFOFILE, txt, 30); 

	if(fs == 0)
	{
		sa = strtol(txt, NULL, 10); 
		return(sa);
	}
	return(fs);
}



/* Write some text to the windvol device */ 
int WVOL_Write(char *txt)
{
	int fd, ws;
	fd = open( DEVICENAME, O_WRONLY); 
	if(fd == -1)
	{
		perror(DEVICENAME);
		return(-1);
	}

	ws = write(fd, txt, strlen(txt));
	close(fd);
	return(ws);
}

int WVOL_Read(char *txt, size_t txtmax)
{
	return WVOL_ReadFile(DEVICENAME, txt, txtmax); 
} 


/* Turn the signals on for us. */ 
int WVOL_TurnOn()
{
	int ws; 
	char txt[80];
	snprintf(txt, 80, "P%u\n", getpid()); 
	ws = WVOL_Write(txt);
	return(ws);
}

/* Turn the signals on for us. */ 
int WVOL_TurnOff()
{
	int ws; 
	ws = WVOL_Write("P0");
	return(ws);
}


long prev_ns = 0;

static void on_cl(int sig, siginfo_t *siginfo, void *c)
{
	struct timespec tnow;
	long dt;

	clock_gettime(CLOCK_REALTIME, &tnow); 

	dt = tnow.tv_nsec - prev_ns;
	if(dt < 0) { dt += 1000000000; } 

	printf("%0ld: Got a 2/3 m hit\n", dt); 
	prev_ns = tnow.tv_nsec;

	return;
}



int G_rflag;

void sigintcatcher(int sig, siginfo_t *siginfo, void *c)
{
	printf("Caught SIGINT, will exit.\n" ); 
	G_rflag = 1;
	return;
}

int main(int argc, char **argv)
{
	struct sigaction sact;
	char txt[40];
	int sig_clwcv;

	sig_clwcv = WVOL_FindSignal();
	printf("Set signal %d\n", sig_clwcv); 

	memset(&sact, 0, sizeof(struct sigaction)); 
	sact.sa_sigaction = sigintcatcher;
	sact.sa_flags = SA_SIGINFO | SA_RESTART; 
	sigaction(SIGINT, &sact, 0);

	memset(&sact, 0, sizeof(struct sigaction)); 
	sact.sa_sigaction = on_cl;
	sact.sa_flags = SA_SIGINFO | SA_RESTART; 
	sigaction(sig_clwcv, &sact, 0);

	G_rflag = 0;

	WVOL_Write("D");
	WVOL_Write("T");
	WVOL_TurnOff();
	WVOL_TurnOn();
	WVOL_Write("E");

	for(;;)
	{
		/* In the loop we can look at the signals having been sent. */

		if(G_rflag) break;
	}

	WVOL_Write("D");
	WVOL_TurnOff();

	WVOL_Read(txt, 40);
	printf("Read count = %s\n", txt); 
	/* Output looks like <<ead count = T:30 C:0 I:1 P:0>> here */ 


	return(0);
}



